import os
import pytest
from pathlib import Path
import tornado

from .utils import expected_http_error

from nbformat import writes
from nbformat.v4 import (new_notebook,
                         new_markdown_cell, new_code_cell,
                         new_output)


@pytest.fixture(params=[
    [False, ['å b']],
    [False, ['å b', 'ç. d']],
    [True, ['.å b']],
    [True, ['å b', '.ç d']]
])
def maybe_hidden(request):
    return request.param


async def fetch_expect_200(jp_fetch, *path_parts):
    r = await jp_fetch('files', *path_parts, method='GET')
    assert (r.body.decode() == path_parts[-1]), (path_parts, r.body)


async def fetch_expect_404(jp_fetch, *path_parts):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch('files', *path_parts, method='GET')
    assert expected_http_error(e, 404), [path_parts, e]


async def test_hidden_files(jp_fetch, jp_serverapp, jp_root_dir, maybe_hidden):
    is_hidden, path_parts = maybe_hidden
    path = Path(jp_root_dir, *path_parts)
    path.mkdir(parents=True, exist_ok=True)

    foos = ['foo', '.foo']
    for foo in foos:
        (path / foo).write_text(foo)

    if is_hidden:
        for foo in foos:
            await fetch_expect_404(jp_fetch, *path_parts, foo)
    else:
        await fetch_expect_404(jp_fetch, *path_parts, '.foo')
        await fetch_expect_200(jp_fetch, *path_parts, 'foo')

    jp_serverapp.contents_manager.allow_hidden = True

    for foo in foos:
        await fetch_expect_200(jp_fetch, *path_parts, foo)


async def test_contents_manager(jp_fetch, jp_serverapp, jp_root_dir):
    """make sure ContentsManager returns right files (ipynb, bin, txt)."""
    nb = new_notebook(
        cells=[
            new_markdown_cell(u'Created by test ³'),
            new_code_cell("print(2*6)", outputs=[
                new_output("stream", text="12"),
            ])
        ]
    )
    jp_root_dir.joinpath('testnb.ipynb').write_text(writes(nb, version=4), encoding='utf-8')
    jp_root_dir.joinpath('test.bin').write_bytes(b'\xff' + os.urandom(5))
    jp_root_dir.joinpath('test.txt').write_text('foobar')

    r = await jp_fetch(
        'files/testnb.ipynb',
        method='GET'
    )
    assert r.code == 200
    assert 'print(2*6)' in r.body.decode('utf-8')

    r = await jp_fetch(
        'files/test.bin',
        method='GET'
    )
    assert r.code == 200
    assert r.headers['content-type'] == 'application/octet-stream'
    assert r.body[:1] == b'\xff'
    assert len(r.body) == 6

    r = await jp_fetch(
        'files/test.txt',
        method='GET'
    )
    assert r.code == 200
    assert r.headers['content-type'] == 'text/plain; charset=UTF-8'
    assert r.body.decode() == 'foobar'


async def test_download(jp_fetch, jp_serverapp, jp_root_dir):
    text = 'hello'
    jp_root_dir.joinpath('test.txt').write_text(text)

    r = await jp_fetch(
        'files', 'test.txt',
        method='GET'
    )
    disposition = r.headers.get('Content-Disposition', '')
    assert 'attachment' not in disposition

    r = await jp_fetch(
        'files', 'test.txt',
        method='GET',
        params={'download': True}
    )
    disposition = r.headers.get('Content-Disposition', '')
    assert 'attachment' in disposition
    assert "filename*=utf-8''test.txt" in disposition


async def test_old_files_redirect(jp_fetch, jp_serverapp, jp_root_dir):
    """pre-2.0 'files/' prefixed links are properly redirected"""
    jp_root_dir.joinpath('files').mkdir(parents=True, exist_ok=True)
    jp_root_dir.joinpath('sub', 'files').mkdir(parents=True, exist_ok=True)

    for prefix in ('', 'sub'):
        jp_root_dir.joinpath(prefix, 'files', 'f1.txt').write_text(prefix + '/files/f1')
        jp_root_dir.joinpath(prefix, 'files', 'f2.txt').write_text(prefix + '/files/f2')
        jp_root_dir.joinpath(prefix, 'f2.txt').write_text(prefix + '/f2')
        jp_root_dir.joinpath(prefix, 'f3.txt').write_text(prefix + '/f3')

    # These depend on the tree handlers
    #
    # def test_download(self):
    #     rootdir = self.root_dir

    #     text = 'hello'
    #     with open(pjoin(rootdir, 'test.txt'), 'w') as f:
    #         f.write(text)

    #     r = self.request('GET', 'files/test.txt')
    #     disposition = r.headers.get('Content-Disposition', '')
    #     self.assertNotIn('attachment', disposition)

    #     r = self.request('GET', 'files/test.txt?download=1')
    #     disposition = r.headers.get('Content-Disposition', '')
    #     self.assertIn('attachment', disposition)
    #     self.assertIn("filename*=utf-8''test.txt", disposition)



"""A mock extension named `mock1` for testing purposes.
"""


# by the test functions.
def _jupyter_server_extension_paths():
    return [
        {
            'module': 'tests.extension.mockextensions.mock1'
        }
    ]


def _load_jupyter_server_extension(serverapp):
    serverapp.mockI = True
    serverapp.mock_shared = 'I'



import pytest


@pytest.fixture
def jp_server_config(jp_template_dir):
    return {
            "ServerApp": {
                "jpserver_extensions": {
                    "tests.extension.mockextensions": True
                }
            },
            "MockExtensionApp": {
                "template_paths": [
                    str(jp_template_dir)
                ]
            }
        }


async def test_handler(jp_fetch):
    r = await jp_fetch(
        'mock',
        method='GET'
    )
    assert r.code == 200
    assert r.body.decode() == 'mock trait'


async def test_handler_template(jp_fetch, mock_template):
    r = await jp_fetch(
        'mock_template',
        method='GET'
    )
    assert r.code == 200


@pytest.mark.parametrize(
    'jp_server_config',
    [
        {
            "ServerApp": {
                "jpserver_extensions": {
                    "tests.extension.mockextensions": True
                }
            },
            "MockExtensionApp": {
                # Change a trait in the MockExtensionApp using
                # the following config value.
                "mock_trait": "test mock trait"
            }
        }
    ]
)
async def test_handler_setting(jp_fetch, jp_server_config):
    # Test that the extension trait was picked up by the webapp.
    r = await jp_fetch(
        'mock',
        method='GET'
    )
    assert r.code == 200
    assert r.body.decode() == 'test mock trait'


@pytest.mark.parametrize(
    'jp_argv', (['--MockExtensionApp.mock_trait=test mock trait'],)
)
async def test_handler_argv(jp_fetch, jp_argv):
    # Test that the extension trait was picked up by the webapp.
    r = await jp_fetch(
        'mock',
        method='GET'
    )
    assert r.code == 200
    assert r.body.decode() == 'test mock trait'


@pytest.mark.parametrize(
    'jp_server_config,jp_base_url',
    [
        (
            {
                "ServerApp": {
                    "jpserver_extensions": {
                        "tests.extension.mockextensions": True
                    },
                    # Move extension handlers behind a url prefix
                    "base_url": "test_prefix"
                },
                "MockExtensionApp": {
                    # Change a trait in the MockExtensionApp using
                    # the following config value.
                    "mock_trait": "test mock trait"
                }
            },
            '/test_prefix/'
        )
    ]
)
async def test_base_url(jp_fetch, jp_server_config, jp_base_url):
    # Test that the extension's handlers were properly prefixed
    r = await jp_fetch(
        'mock',
        method='GET'
    )
    assert r.code == 200
    assert r.body.decode() == 'test mock trait'

    # Test that the static namespace was prefixed by base_url
    r = await jp_fetch(
        'static', 'mockextension', 'mock.txt',
        method='GET'
    )
    assert r.code == 200
    body = r.body.decode()
    assert "mock static content" in body



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import sys
import json
import pytest
import shutil
import urllib.parse

from binascii import hexlify

import tornado
from tornado.escape import url_escape
import jupyter_core.paths
import nbformat
from traitlets.config import Config

from jupyter_server.extension import serverextension
from jupyter_server.serverapp import ServerApp
from jupyter_server.utils import url_path_join
from jupyter_server.services.contents.filemanager import FileContentsManager
from jupyter_server.services.contents.largefilemanager import LargeFileManager


# List of dependencies needed for this plugin.
pytest_plugins = [
    "pytest_tornasync",
    # Once the chunk below moves to Jupyter Core, we'll uncomment
    # This plugin and use the fixtures directly from Jupyter Core.
    # "jupyter_core.pytest_plugin"
]

# ============ Move to Jupyter Core =============

def mkdir(tmp_path, *parts):
    path = tmp_path.joinpath(*parts)
    if not path.exists():
        path.mkdir(parents=True)
    return path


@pytest.fixture
def jp_home_dir(tmp_path):
    """Provides a temporary HOME directory value."""
    return mkdir(tmp_path, "home")


@pytest.fixture
def jp_data_dir(tmp_path):
    """Provides a temporary Jupyter data dir directory value."""
    return mkdir(tmp_path, "data")


@pytest.fixture
def jp_config_dir(tmp_path):
    """Provides a temporary Jupyter config dir directory value."""
    return mkdir(tmp_path, "config")


@pytest.fixture
def jp_runtime_dir(tmp_path):
    """Provides a temporary Jupyter runtime dir directory value."""
    return mkdir(tmp_path, "runtime")


@pytest.fixture
def jp_system_jupyter_path(tmp_path):
    """Provides a temporary Jupyter system path value."""
    return mkdir(tmp_path, "share", "jupyter")


@pytest.fixture
def jp_env_jupyter_path(tmp_path):
    """Provides a temporary Jupyter env system path value."""
    return mkdir(tmp_path, "env", "share", "jupyter")


@pytest.fixture
def jp_system_config_path(tmp_path):
    """Provides a temporary Jupyter config path value."""
    return mkdir(tmp_path, "etc", "jupyter")


@pytest.fixture
def jp_env_config_path(tmp_path):
    """Provides a temporary Jupyter env config path value."""
    return mkdir(tmp_path, "env", "etc", "jupyter")


@pytest.fixture
def jp_environ(
    monkeypatch,
    tmp_path,
    jp_home_dir,
    jp_data_dir,
    jp_config_dir,
    jp_runtime_dir,
    jp_system_jupyter_path,
    jp_system_config_path,
    jp_env_jupyter_path,
    jp_env_config_path,
):
    """Configures a temporary environment based on Jupyter-specific environment variables. """
    monkeypatch.setenv("HOME", str(jp_home_dir))
    monkeypatch.setenv("PYTHONPATH", os.pathsep.join(sys.path))
    # monkeypatch.setenv("JUPYTER_NO_CONFIG", "1")
    monkeypatch.setenv("JUPYTER_CONFIG_DIR", str(jp_config_dir))
    monkeypatch.setenv("JUPYTER_DATA_DIR", str(jp_data_dir))
    monkeypatch.setenv("JUPYTER_RUNTIME_DIR", str(jp_runtime_dir))
    monkeypatch.setattr(
        jupyter_core.paths, "SYSTEM_JUPYTER_PATH", [str(jp_system_jupyter_path)]
    )
    monkeypatch.setattr(jupyter_core.paths, "ENV_JUPYTER_PATH", [str(jp_env_jupyter_path)])
    monkeypatch.setattr(
        jupyter_core.paths, "SYSTEM_CONFIG_PATH", [str(jp_system_config_path)]
    )
    monkeypatch.setattr(jupyter_core.paths, "ENV_CONFIG_PATH", [str(jp_env_config_path)])


# ================= End: Move to Jupyter core ================

# NOTE: This is a temporary fix for Windows 3.8
# We have to override the io_loop fixture with an
# asyncio patch. This will probably be removed in
# the future.
@pytest.fixture
def jp_asyncio_patch():
    """Appropriately configures the event loop policy if running on Windows w/ Python >= 3.8."""
    ServerApp()._init_asyncio_patch()


@pytest.fixture
def io_loop(jp_asyncio_patch):
    """Returns an ioloop instance that includes the asyncio patch for Windows 3.8 platforms."""
    loop = tornado.ioloop.IOLoop()
    loop.make_current()
    yield loop
    loop.clear_current()
    loop.close(all_fds=True)


@pytest.fixture
def jp_server_config():
    """Allows tests to setup their specific configuration values. """
    return {}


@pytest.fixture
def jp_root_dir(tmp_path):
    """Provides a temporary Jupyter root directory value."""
    return mkdir(tmp_path, "root_dir")


@pytest.fixture
def jp_template_dir(tmp_path):
    """Provides a temporary Jupyter templates directory value."""
    return mkdir(tmp_path, "templates")


@pytest.fixture
def jp_argv():
    """Allows tests to setup specific argv values. """
    return []


@pytest.fixture
def jp_extension_environ(jp_env_config_path, monkeypatch):
    """Monkeypatch a Jupyter Extension's config path into each test's environment variable"""
    monkeypatch.setattr(serverextension, "ENV_CONFIG_PATH", [str(jp_env_config_path)])


@pytest.fixture
def jp_http_port(http_server_port):
    """Returns the port value from the http_server_port fixture. """
    return http_server_port[-1]


@pytest.fixture
def jp_nbconvert_templates(jp_data_dir):
    """Setups up a temporary directory consisting of the nbconvert templates."""

    # Get path to nbconvert template directory *before*
    # monkeypatching the paths env variable via the jp_environ fixture.
    possible_paths = jupyter_core.paths.jupyter_path('nbconvert', 'templates')
    nbconvert_path = None
    for path in possible_paths:
        if os.path.exists(path):
            nbconvert_path = path
            break

    nbconvert_target = jp_data_dir / 'nbconvert' / 'templates'

    # copy nbconvert templates to new tmp data_dir.
    if nbconvert_path:
        shutil.copytree(nbconvert_path, str(nbconvert_target))


@pytest.fixture(scope='function')
def jp_configurable_serverapp(
    jp_nbconvert_templates,  # this fixture must preceed jp_environ
    jp_environ,
    jp_server_config,
    jp_argv,
    jp_http_port,
    jp_base_url,
    tmp_path,
    jp_root_dir,
    io_loop,
):
    """Starts a Jupyter Server instance based on
    the provided configuration values.

    The fixture is a factory; it can be called like
    a function inside a unit test. Here's a basic
    example of how use this fixture:

    .. code-block:: python

        def my_test(jp_configurable_serverapp):

            app = jp_configurable_serverapp(...)
            ...
    """
    ServerApp.clear_instance()

    def _configurable_serverapp(
        config=jp_server_config,
        base_url=jp_base_url,
        argv=jp_argv,
        environ=jp_environ,
        http_port=jp_http_port,
        tmp_path=tmp_path,
        root_dir=jp_root_dir,
        **kwargs
    ):
        c = Config(config)
        c.NotebookNotary.db_file = ":memory:"
        token = hexlify(os.urandom(4)).decode("ascii")
        app = ServerApp.instance(
            # Set the log level to debug for testing purposes
            log_level='DEBUG',
            port=http_port,
            port_retries=0,
            open_browser=False,
            root_dir=str(root_dir),
            base_url=base_url,
            config=c,
            allow_root=True,
            token=token,
            **kwargs
        )

        app.init_signal = lambda: None
        app.log.propagate = True
        app.log.handlers = []
        # Initialize app without httpserver
        app.initialize(argv=argv, new_httpserver=False)
        app.log.propagate = True
        app.log.handlers = []
        # Start app without ioloop
        app.start_app()
        return app

    return _configurable_serverapp


@pytest.fixture
def jp_ensure_app_fixture(request):
    """Ensures that the 'app' fixture used by pytest-tornasync
    is set to `jp_web_app`, the Tornado Web Application returned
    by the ServerApp in Jupyter Server, provided by the jp_web_app
    fixture in this module.

    Note, this hardcodes the `app_fixture` option from
    pytest-tornasync to `jp_web_app`. If this value is configured
    to something other than the default, it will raise an exception.
    """
    app_option = request.config.getoption("app_fixture")
    if app_option not in ["app", "jp_web_app"]:
        raise Exception("jp_serverapp requires the `app-fixture` option "
                        "to be set to 'jp_web_app`. Try rerunning the "
                        "current tests with the option `--app-fixture "
                        "jp_web_app`.")
    elif app_option == "app":
        # Manually set the app_fixture to `jp_web_app` if it's
        # not set already.
        request.config.option.app_fixture = "jp_web_app"


@pytest.fixture(scope="function")
def jp_serverapp(
    jp_ensure_app_fixture,
    jp_server_config,
    jp_argv,
    jp_configurable_serverapp
):
    """Starts a Jupyter Server instance based on the established configuration values."""
    app = jp_configurable_serverapp(config=jp_server_config, argv=jp_argv)
    yield app
    app.remove_server_info_file()
    app.remove_browser_open_file()
    app.cleanup_kernels()


@pytest.fixture
def jp_web_app(jp_serverapp):
    """app fixture is needed by pytest_tornasync plugin"""
    return jp_serverapp.web_app


@pytest.fixture
def jp_auth_header(jp_serverapp):
    """Configures an authorization header using the token from the serverapp fixture."""
    return {"Authorization": "token {token}".format(token=jp_serverapp.token)}


@pytest.fixture
def jp_base_url():
    """Returns the base url to use for the test."""
    return "/a%40b/"


@pytest.fixture
def jp_fetch(jp_serverapp, http_server_client, jp_auth_header, jp_base_url):
    """Sends an (asynchronous) HTTP request to a test server.

    The fixture is a factory; it can be called like
    a function inside a unit test. Here's a basic
    example of how use this fixture:

    .. code-block:: python

        async def my_test(jp_fetch):

            response = await jp_fetch("api", "spec.yaml")
            ...
    """
    def client_fetch(*parts, headers={}, params={}, **kwargs):
        # Handle URL strings
        path_url = url_escape(url_path_join(*parts), plus=False)
        base_path_url = url_path_join(jp_base_url, path_url)
        params_url = urllib.parse.urlencode(params)
        url = base_path_url + "?" + params_url
        # Add auth keys to header
        headers.update(jp_auth_header)
        # Make request.
        return http_server_client.fetch(
            url, headers=headers, request_timeout=20, **kwargs
        )
    return client_fetch


@pytest.fixture
def jp_ws_fetch(jp_serverapp, jp_auth_header, jp_http_port, jp_base_url):
    """Sends a websocket request to a test server.

    The fixture is a factory; it can be called like
    a function inside a unit test. Here's a basic
    example of how use this fixture:

    .. code-block:: python

        async def my_test(jp_fetch, jp_ws_fetch):
            # Start a kernel
            r = await jp_fetch(
                'api', 'kernels',
                method='POST',
                body=json.dumps({
                    'name': "python3"
                })
            )
            kid = json.loads(r.body.decode())['id']

            # Open a websocket connection.
            ws = await jp_ws_fetch(
                'api', 'kernels', kid, 'channels'
            )
            ...
    """
    def client_fetch(*parts, headers={}, params={}, **kwargs):
        # Handle URL strings
        path_url = url_escape(url_path_join(*parts), plus=False)
        base_path_url = url_path_join(jp_base_url, path_url)
        urlparts = urllib.parse.urlparse('ws://localhost:{}'.format(jp_http_port))
        urlparts = urlparts._replace(
            path=base_path_url,
            query=urllib.parse.urlencode(params)
        )
        url = urlparts.geturl()
        # Add auth keys to header
        headers.update(jp_auth_header)
        # Make request.
        req = tornado.httpclient.HTTPRequest(
            url,
            headers=jp_auth_header,
            connect_timeout=120
        )
        return tornado.websocket.websocket_connect(req)
    return client_fetch


some_resource = u"The very model of a modern major general"
sample_kernel_json = {
    'argv':['cat', '{connection_file}'],
    'display_name': 'Test kernel',
}
@pytest.fixture
def jp_kernelspecs(jp_data_dir):
    """Configures some sample kernelspecs in the Jupyter data directory."""
    spec_names = ['sample', 'sample 2']
    for name in spec_names:
        sample_kernel_dir = jp_data_dir.joinpath('kernels', name)
        sample_kernel_dir.mkdir(parents=True)
        # Create kernel json file
        sample_kernel_file = sample_kernel_dir.joinpath('kernel.json')
        sample_kernel_file.write_text(json.dumps(sample_kernel_json))
        # Create resources text
        sample_kernel_resources = sample_kernel_dir.joinpath('resource.txt')
        sample_kernel_resources.write_text(some_resource)


@pytest.fixture(params=[True, False])
def jp_contents_manager(request, tmp_path):
    """Returns a FileContentsManager instance based on the use_atomic_writing parameter value."""
    return FileContentsManager(root_dir=str(tmp_path), use_atomic_writing=request.param)


@pytest.fixture
def jp_large_contents_manager(tmp_path):
    """Returns a LargeFileManager instance."""
    return LargeFileManager(root_dir=str(tmp_path))


@pytest.fixture
def jp_create_notebook(jp_root_dir):
    """Creates a notebook in the test's home directory."""
    def inner(nbpath):
        nbpath = jp_root_dir.joinpath(nbpath)
        # Check that the notebook has the correct file extension.
        if nbpath.suffix != '.ipynb':
            raise Exception("File extension for notebook must be .ipynb")
        # If the notebook path has a parent directory, make sure it's created.
        parent = nbpath.parent
        parent.mkdir(parents=True, exist_ok=True)
        # Create a notebook string and write to file.
        nb = nbformat.v4.new_notebook()
        nbtext = nbformat.writes(nb, version=4)
        nbpath.write_text(nbtext)
    return inner



import pytest
from jupyter_server.extension.utils import validate_extension


# Use ServerApps environment because it monkeypatches
# jupyter_core.paths and provides a config directory
# that's not cross contaminating the user config directory.
pytestmark = pytest.mark.usefixtures("jp_environ")


def test_validate_extension():
    # enabled at sys level
    assert validate_extension('tests.extension.mockextensions.mockext_sys')
    # enabled at sys, disabled at user
    assert validate_extension('tests.extension.mockextensions.mockext_both')
    # enabled at user
    assert validate_extension('tests.extension.mockextensions.mockext_user')
    # enabled at Python
    assert validate_extension('tests.extension.mockextensions.mockext_py')


import io
import os
import stat
import functools
import decorator
import pytest
import sys
from ipython_genutils.testing.decorators import skip_win32 as _skip_win32

from jupyter_server.services.contents.fileio import atomic_writing


@functools.wraps(_skip_win32)
def skip_win32(f):
    # Patches the "skip_win32" method to allow pytest fixtures
    # in methods wrapped by this decorator.
    def inner(f, *args, **kwargs):
        decorated_f = _skip_win32(f)
        return decorated_f(*args, **kwargs)
    return decorator.decorator(inner, f)


umask = 0


def test_atomic_writing(tmp_path):
    class CustomExc(Exception): pass

    f1 = tmp_path / 'penguin'
    f1.write_text('Before')

    if os.name != 'nt':
        os.chmod(str(f1), 0o701)
        orig_mode = stat.S_IMODE(os.stat(str(f1)).st_mode)

    f2 = tmp_path / 'flamingo'
    try:
        os.symlink(str(f1), str(f2))
        have_symlink = True
    except (AttributeError, NotImplementedError, OSError):
        # AttributeError: Python doesn't support it
        # NotImplementedError: The system doesn't support it
        # OSError: The user lacks the privilege (Windows)
        have_symlink = False

    with pytest.raises(CustomExc):
        with atomic_writing(str(f1)) as f:
            f.write('Failing write')
            raise CustomExc

    with io.open(str(f1), 'r') as f:
        assert f.read() == 'Before'
    
    with atomic_writing(str(f1)) as f:
        f.write('Overwritten')

    with io.open(str(f1), 'r') as f:
        assert f.read() == 'Overwritten'

    if os.name != 'nt':
        mode = stat.S_IMODE(os.stat(str(f1)).st_mode)
        assert mode == orig_mode

    if have_symlink:
        # Check that writing over a file preserves a symlink
        with atomic_writing(str(f2)) as f:
            f.write(u'written from symlink')
        
        with io.open(str(f1), 'r') as f:
            assert f.read() == 'written from symlink'


@pytest.fixture
def handle_umask():
    global umask
    umask = os.umask(0)
    os.umask(umask)
    yield
    os.umask(umask)


@pytest.mark.skipif(sys.platform.startswith('win'), reason="Windows")
def test_atomic_writing_umask(handle_umask, tmp_path):

    os.umask(0o022)
    f1 = str(tmp_path / '1')
    with atomic_writing(f1) as f:
        f.write('1')
    mode = stat.S_IMODE(os.stat(f1).st_mode)
    assert mode == 0o644

    os.umask(0o057)
    f2 = str(tmp_path / '2')

    with atomic_writing(f2) as f:
        f.write('2')

    mode = stat.S_IMODE(os.stat(f2).st_mode)
    assert mode == 0o620


def test_atomic_writing_newlines(tmp_path):
    path = str(tmp_path / 'testfile')

    lf = u'a\nb\nc\n'
    plat = lf.replace(u'\n', os.linesep)
    crlf = lf.replace(u'\n', u'\r\n')

    # test default
    with io.open(path, 'w') as f:
        f.write(lf)
    with io.open(path, 'r', newline='') as f:
        read = f.read()
    assert read == plat

    # test newline=LF
    with io.open(path, 'w', newline='\n') as f:
        f.write(lf)
    with io.open(path, 'r', newline='') as f:
        read = f.read()
    assert read == lf

    # test newline=CRLF
    with atomic_writing(str(path), newline='\r\n') as f:
        f.write(lf)
    with io.open(path, 'r', newline='') as f:
        read = f.read()
    assert read == crlf

    # test newline=no convert
    text = u'crlf\r\ncr\rlf\n'
    with atomic_writing(str(path), newline='') as f:
        f.write(text)
    with io.open(path, 'r', newline='') as f:
        read = f.read()
    assert read == text


import sys
import json
import pathlib
import pytest
from urllib.parse import ParseResult, urlunparse

import tornado

from nbformat import writes, from_dict
from nbformat.v4 import (
    new_notebook, new_markdown_cell,
)

from jupyter_server.utils import url_path_join

from base64 import encodebytes, decodebytes

from ...utils import expected_http_error


def notebooks_only(dir_model):
    return [nb for nb in dir_model['content'] if nb['type']=='notebook']

def dirs_only(dir_model):
    return [x for x in dir_model['content'] if x['type']=='directory']


dirs = [
    ('', 'inroot'),
    ('Directory with spaces in', 'inspace'),
    (u'unicodé', 'innonascii'),
    ('foo', 'a'),
    ('foo', 'b'),
    ('foo', 'name with spaces'),
    ('foo', u'unicodé'),
    ('foo/bar', 'baz'),
    ('ordering', 'A'),
    ('ordering', 'b'),
    ('ordering', 'C'),
    (u'å b', u'ç d'),
]


@pytest.fixture(params=["FileContentsManager", "AsyncFileContentsManager"])
def jp_argv(request):
    return ["--ServerApp.contents_manager_class=jupyter_server.services.contents.filemanager." + request.param]


@pytest.fixture
def contents_dir(tmp_path, jp_serverapp):
    return tmp_path / jp_serverapp.root_dir


@pytest.fixture
def contents(contents_dir):
    # Create files in temporary directory
    for d, name in dirs:
        p = contents_dir / d
        p.mkdir(parents=True, exist_ok=True)

        # Create a notebook
        nb = writes(new_notebook(), version=4)
        nbname = p.joinpath('{}.ipynb'.format(name))
        nbname.write_text(nb, encoding='utf-8')

        # Create a text file
        txt = '{} text file'.format(name)
        txtname = p.joinpath('{}.txt'.format(name))
        txtname.write_text(txt, encoding='utf-8')

        # Create a random blob
        blob = name.encode('utf-8') + b'\xFF'
        blobname = p.joinpath('{}.blob'.format(name))
        blobname.write_bytes(blob)


@pytest.fixture
def folders():
    return list(set(item[0] for item in dirs))


@pytest.mark.parametrize('path,name', dirs)
async def test_list_notebooks(jp_fetch, contents, path, name):
    response = await jp_fetch(
        'api', 'contents', path,
        method='GET',
    )
    data = json.loads(response.body.decode())
    nbs = notebooks_only(data)
    assert len(nbs) > 0
    assert name+'.ipynb' in [n['name'] for n in nbs]
    assert url_path_join(path, name+'.ipynb') in [n['path'] for n in nbs]


@pytest.mark.parametrize('path,name', dirs)
async def test_get_dir_no_contents(jp_fetch, contents, path, name):
    response = await jp_fetch(
        'api', 'contents', path,
        method='GET',
        params=dict(
            content='0',
        )
    )
    model = json.loads(response.body.decode())
    assert model['path'] == path
    assert model['type'] == 'directory'
    assert 'content' in model
    assert model['content'] is None


async def test_list_nonexistant_dir(jp_fetch, contents):
    with pytest.raises(tornado.httpclient.HTTPClientError):
        await jp_fetch(
            'api', 'contents', 'nonexistant',
            method='GET',
        )


@pytest.mark.parametrize('path,name', dirs)
async def test_get_nb_contents(jp_fetch, contents, path, name):
    nbname = name+'.ipynb'
    nbpath = (path + '/' + nbname).lstrip('/')
    r = await jp_fetch(
        'api', 'contents', nbpath,
        method='GET',
        params=dict(content='1')
    )
    model = json.loads(r.body.decode())
    assert model['name'] == nbname
    assert model['path'] == nbpath
    assert model['type'] == 'notebook'
    assert 'content' in model
    assert model['format'] == 'json'
    assert 'metadata' in model['content']
    assert isinstance(model['content']['metadata'], dict)


@pytest.mark.parametrize('path,name', dirs)
async def test_get_nb_no_contents(jp_fetch, contents, path, name):
    nbname = name+'.ipynb'
    nbpath = (path + '/' + nbname).lstrip('/')
    r = await jp_fetch(
        'api', 'contents', nbpath,
        method='GET',
        params=dict(content='0')
    )
    model = json.loads(r.body.decode())
    assert model['name'] == nbname
    assert model['path'] == nbpath
    assert model['type'] == 'notebook'
    assert 'content' in model
    assert model['content'] is None


async def test_get_nb_invalid(contents_dir, jp_fetch, contents):
    nb = {
        'nbformat': 4,
        'metadata': {},
        'cells': [{
            'cell_type': 'wrong',
            'metadata': {},
        }],
    }
    nbpath = u'å b/Validate tést.ipynb'
    (contents_dir / nbpath).write_text(json.dumps(nb))
    r = await jp_fetch(
        'api', 'contents', nbpath,
        method='GET',
    )
    model = json.loads(r.body.decode())
    assert model['path'] == nbpath
    assert model['type'] == 'notebook'
    assert 'content' in model
    assert 'message' in model
    assert 'validation failed' in model['message'].lower()


async def test_get_contents_no_such_file(jp_fetch):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch(
            'api', 'contents', 'foo/q.ipynb',
            method='GET',
        )
    assert e.value.code == 404


@pytest.mark.parametrize('path,name', dirs)
async def test_get_text_file_contents(jp_fetch, contents, path, name):
    txtname = name+'.txt'
    txtpath = (path + '/' + txtname).lstrip('/')
    r = await jp_fetch(
        'api', 'contents', txtpath,
        method='GET',
        params=dict(content='1')
    )
    model = json.loads(r.body.decode())
    assert model['name'] == txtname
    assert model['path'] == txtpath
    assert 'content' in model
    assert model['format'] == 'text'
    assert model['type'] == 'file'
    assert model['content'] == '{} text file'.format(name)

    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch(
            'api', 'contents', 'foo/q.txt',
            method='GET',
        )
    assert expected_http_error(e, 404)

    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch(
            'api', 'contents', 'foo/bar/baz.blob',
            method='GET',
            params=dict(
                type='file',
                format='text'
            )
        )
    assert expected_http_error(e, 400)



@pytest.mark.parametrize('path,name', dirs)
async def test_get_binary_file_contents(jp_fetch, contents, path, name):
    blobname = name+'.blob'
    blobpath = (path + '/' + blobname).lstrip('/')
    r = await jp_fetch(
        'api', 'contents', blobpath,
        method='GET',
        params=dict(content='1')
    )
    model = json.loads(r.body.decode())
    assert model['name'] == blobname
    assert model['path'] == blobpath
    assert 'content' in model
    assert model['format'] == 'base64'
    assert model['type'] == 'file'
    data_out = decodebytes(model['content'].encode('ascii'))
    data_in = name.encode('utf-8') + b'\xFF'
    assert data_in == data_out

    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch(
            'api', 'contents', 'foo/q.txt',
            method='GET',
        )
    assert expected_http_error(e, 404)


async def test_get_bad_type(jp_fetch, contents):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        path = 'unicodé'
        type = 'file'
        await jp_fetch(
            'api', 'contents', path,
            method='GET',
            params=dict(type=type) # This should be a directory, and thus throw and error
        )
    assert expected_http_error(e, 400, '%s is a directory, not a %s' % (path, type))

    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        path = 'unicodé/innonascii.ipynb'
        type = 'directory'
        await jp_fetch(
            'api', 'contents', path,
            method='GET',
            params=dict(type=type) # This should be a file, and thus throw and error
        )
    assert expected_http_error(e, 400, '%s is not a directory' % path)

@pytest.fixture
def _check_created(jp_base_url):
    def _inner(r, contents_dir, path, name, type='notebook'):
        fpath = path+'/'+name
        assert r.code == 201
        location = jp_base_url + 'api/contents/' + tornado.escape.url_escape(fpath, plus=False)
        assert r.headers['Location'] == location
        model = json.loads(r.body.decode())
        assert model['name'] == name
        assert model['path'] == fpath
        assert model['type'] == type
        path = contents_dir + '/' + fpath
        if type == 'directory':
            assert pathlib.Path(path).is_dir()
        else:
            assert pathlib.Path(path).is_file()
    return _inner

async def test_create_untitled(jp_fetch, contents, contents_dir, _check_created):
    path = 'å b'
    name = 'Untitled.ipynb'
    r = await jp_fetch(
        'api', 'contents', path,
        method='POST',
        body=json.dumps({'ext': '.ipynb'})
    )
    _check_created(r, str(contents_dir), path, name, type='notebook')

    name = 'Untitled1.ipynb'
    r = await jp_fetch(
        'api', 'contents', path,
        method='POST',
        body=json.dumps({'ext': '.ipynb'})
    )
    _check_created(r, str(contents_dir), path, name, type='notebook')

    path = 'foo/bar'
    name = 'Untitled.ipynb'
    r = await jp_fetch(
        'api', 'contents', path,
        method='POST',
        body=json.dumps({'ext': '.ipynb'})
    )
    _check_created(r, str(contents_dir), path, name, type='notebook')


async def test_create_untitled_txt(jp_fetch, contents, contents_dir, _check_created):
    name = 'untitled.txt'
    path = 'foo/bar'
    r = await jp_fetch(
        'api', 'contents', path,
        method='POST',
        body=json.dumps({'ext': '.txt'})
    )
    _check_created(r, str(contents_dir), path, name, type='file')

    r = await jp_fetch(
        'api', 'contents', path, name,
        method='GET'
    )
    model = json.loads(r.body.decode())
    assert model['type'] == 'file'
    assert model['format'] == 'text'
    assert model['content'] == ''


async def test_upload(jp_fetch, contents, contents_dir, _check_created):
    nb = new_notebook()
    nbmodel = {'content': nb, 'type': 'notebook'}
    path = 'å b'
    name = 'Upload tést.ipynb'
    r = await jp_fetch(
        'api', 'contents', path, name,
        method='PUT',
        body=json.dumps(nbmodel)
    )
    _check_created(r, str(contents_dir), path, name)


async def test_mkdir_untitled(jp_fetch, contents, contents_dir, _check_created):
    name = 'Untitled Folder'
    path = 'å b'
    r = await jp_fetch(
        'api', 'contents', path,
        method='POST',
        body=json.dumps({'type': 'directory'})
    )
    _check_created(r, str(contents_dir), path, name, type='directory')

    name = 'Untitled Folder 1'
    r = await jp_fetch(
        'api', 'contents', path,
        method='POST',
        body=json.dumps({'type': 'directory'})
    )
    _check_created(r, str(contents_dir), path, name, type='directory')

    name = 'Untitled Folder'
    path = 'foo/bar'
    r = await jp_fetch(
        'api', 'contents', path,
        method='POST',
        body=json.dumps({'type': 'directory'})
    )
    _check_created(r, str(contents_dir), path, name, type='directory')


async def test_mkdir(jp_fetch, contents, contents_dir, _check_created):
    name = 'New ∂ir'
    path = 'å b'
    r = await jp_fetch(
        'api', 'contents', path, name,
        method='PUT',
        body=json.dumps({'type': 'directory'})
    )
    _check_created(r, str(contents_dir), path, name, type='directory')


async def test_mkdir_hidden_400(jp_fetch):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch(
            'api', 'contents', 'å b/.hidden',
            method='PUT',
            body=json.dumps({'type': 'directory'})
        )
    assert expected_http_error(e, 400)


async def test_upload_txt(jp_fetch, contents, contents_dir, _check_created):
    body = 'ünicode téxt'
    model = {
        'content' : body,
        'format'  : 'text',
        'type'    : 'file',
    }
    path = 'å b'
    name = 'Upload tést.txt'
    await jp_fetch(
        'api', 'contents', path, name,
        method='PUT',
        body=json.dumps(model)
    )

    # check roundtrip
    r = await jp_fetch(
        'api', 'contents', path, name,
        method='GET'
    )
    model = json.loads(r.body.decode())
    assert model['type'] == 'file'
    assert model['format'] == 'text'
    assert model['path'] == path+'/'+name
    assert model['content'] == body


async def test_upload_b64(jp_fetch, contents, contents_dir, _check_created):
    body = b'\xFFblob'
    b64body = encodebytes(body).decode('ascii')
    model = {
        'content' : b64body,
        'format'  : 'base64',
        'type'    : 'file',
    }
    path = 'å b'
    name = 'Upload tést.blob'
    await jp_fetch(
        'api', 'contents', path, name,
        method='PUT',
        body=json.dumps(model)
    )
    # check roundtrip
    r = await jp_fetch(
        'api', 'contents', path, name,
        method='GET'
    )
    model = json.loads(r.body.decode())
    assert model['type'] == 'file'
    assert model['path'] == path+'/'+name
    assert model['format'] == 'base64'
    decoded = decodebytes(model['content'].encode('ascii'))
    assert decoded == body


async def test_copy(jp_fetch, contents, contents_dir, _check_created):
    path = 'å b'
    name = 'ç d.ipynb'
    copy = 'ç d-Copy1.ipynb'
    r = await jp_fetch(
        'api', 'contents', path,
        method='POST',
        body=json.dumps({'copy_from': path+'/'+name})
    )
    _check_created(r, str(contents_dir), path, copy, type='notebook')

    # Copy the same file name
    copy2 = 'ç d-Copy2.ipynb'
    r = await jp_fetch(
        'api', 'contents', path,
        method='POST',
        body=json.dumps({'copy_from': path+'/'+name})
    )
    _check_created(r, str(contents_dir), path, copy2, type='notebook')

    # copy a copy.
    copy3 = 'ç d-Copy3.ipynb'
    r = await jp_fetch(
        'api', 'contents', path,
        method='POST',
        body=json.dumps({'copy_from': path+'/'+copy2})
    )
    _check_created(r, str(contents_dir), path, copy3, type='notebook')


async def test_copy_path(jp_fetch, contents, contents_dir, _check_created):
    path1 = 'foo'
    path2 = 'å b'
    name = 'a.ipynb'
    copy = 'a-Copy1.ipynb'
    r = await jp_fetch(
        'api', 'contents', path2,
        method='POST',
        body=json.dumps({'copy_from': path1+'/'+name})
    )
    _check_created(r, str(contents_dir), path2, name, type='notebook')

    r = await jp_fetch(
        'api', 'contents', path2,
        method='POST',
        body=json.dumps({'copy_from': path1+'/'+name})
    )
    _check_created(r, str(contents_dir), path2, copy, type='notebook')


async def test_copy_put_400(jp_fetch, contents, contents_dir, _check_created):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch(
            'api', 'contents', 'å b/cøpy.ipynb',
            method='PUT',
            body=json.dumps({'copy_from': 'å b/ç d.ipynb'})
        )
    assert expected_http_error(e, 400)


async def test_copy_dir_400(jp_fetch, contents, contents_dir, _check_created):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch(
            'api', 'contents', 'foo',
            method='POST',
            body=json.dumps({'copy_from': 'å b'})
        )
    assert expected_http_error(e, 400)


@pytest.mark.parametrize('path,name', dirs)
async def test_delete(jp_fetch, contents, contents_dir, path, name, _check_created):
    nbname = name+'.ipynb'
    nbpath = (path + '/' + nbname).lstrip('/')
    r = await jp_fetch(
        'api', 'contents', nbpath,
        method='DELETE',
    )
    assert r.code == 204


async def test_delete_dirs(jp_fetch, contents, folders):
    # Iterate over folders
    for name in sorted(folders + ['/'], key=len, reverse=True):
        r = await jp_fetch(
            'api', 'contents', name,
            method='GET'
        )
        # Get JSON blobs for each content.
        listing = json.loads(r.body.decode())['content']
        # Delete all content
        for model in listing:
            await jp_fetch(
                'api', 'contents', model['path'],
                method='DELETE'
            )
    # Make sure all content has been deleted.
    r = await jp_fetch(
        'api', 'contents',
        method='GET'
    )
    model = json.loads(r.body.decode())
    assert model['content'] == []


@pytest.mark.skipif(sys.platform == 'win32', reason="Disabled deleting non-empty dirs on Windows")
async def test_delete_non_empty_dir(jp_fetch, contents):
    # Delete a folder
    await jp_fetch(
        'api', 'contents', 'å b',
        method='DELETE'
    )
    # Check that the folder was been deleted.
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch(
            'api', 'contents', 'å b',
            method='GET'
        )
    assert expected_http_error(e, 404)


async def test_rename(jp_fetch, jp_base_url, contents, contents_dir):
    path = 'foo'
    name = 'a.ipynb'
    new_name = 'z.ipynb'
    # Rename the file
    r = await jp_fetch(
        'api', 'contents', path, name,
        method='PATCH',
        body=json.dumps({'path': path+'/'+new_name})
    )
    fpath = path+'/'+new_name
    assert r.code == 200
    location = url_path_join(jp_base_url, 'api/contents/', fpath)
    assert r.headers['Location'] == location
    model = json.loads(r.body.decode())
    assert model['name'] == new_name
    assert model['path'] == fpath
    fpath = str(contents_dir / fpath)
    assert pathlib.Path(fpath).is_file()

    # Check that the files have changed
    r = await jp_fetch(
        'api', 'contents', path,
        method='GET'
    )
    listing = json.loads(r.body.decode())
    nbnames = [name['name'] for name in listing['content']]
    assert 'z.ipynb' in nbnames
    assert 'a.ipynb' not in nbnames


async def test_checkpoints_follow_file(jp_fetch, contents):
    path = 'foo'
    name = 'a.ipynb'

    # Read initial file.
    r = await jp_fetch(
        'api', 'contents', path, name,
        method='GET'
    )
    model = json.loads(r.body.decode())

    # Create a checkpoint of initial state
    r = await jp_fetch(
        'api', 'contents', path, name, 'checkpoints',
        method='POST',
        allow_nonstandard_methods=True
    )
    cp1 = json.loads(r.body.decode())

    # Modify file and save.
    nbcontent = model['content']
    nb = from_dict(nbcontent)
    hcell = new_markdown_cell('Created by test')
    nb.cells.append(hcell)
    nbmodel = {'content': nb, 'type': 'notebook'}
    await jp_fetch(
        'api', 'contents', path, name,
        method='PUT',
        body=json.dumps(nbmodel)
    )

    # List checkpoints
    r = await jp_fetch(
        'api', 'contents', path, name, 'checkpoints',
        method='GET',
    )
    cps = json.loads(r.body.decode())
    assert cps == [cp1]

    r = await jp_fetch(
        'api', 'contents', path, name,
        method='GET'
    )
    model = json.loads(r.body.decode())
    nbcontent = model['content']
    nb = from_dict(nbcontent)
    assert nb.cells[0].source == "Created by test"


async def test_rename_existing(jp_fetch, contents):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        path = 'foo'
        name = 'a.ipynb'
        new_name = 'b.ipynb'
        # Rename the file
        await jp_fetch(
            'api', 'contents', path, name,
            method='PATCH',
            body=json.dumps({'path': path+'/'+new_name})
        )
    assert expected_http_error(e, 409)


async def test_save(jp_fetch, contents):
    r = await jp_fetch(
        'api', 'contents', 'foo/a.ipynb',
        method='GET'
    )
    model = json.loads(r.body.decode())
    nbmodel = model['content']
    nb = from_dict(nbmodel)
    nb.cells.append(new_markdown_cell('Created by test ³'))
    nbmodel = {'content': nb, 'type': 'notebook'}
    await jp_fetch(
        'api', 'contents', 'foo/a.ipynb',
        method='PUT',
        body=json.dumps(nbmodel)
    )
    # Round trip.
    r = await jp_fetch(
        'api', 'contents', 'foo/a.ipynb',
        method='GET'
    )
    model = json.loads(r.body.decode())
    newnb = from_dict(model['content'])
    assert newnb.cells[0].source == 'Created by test ³'


async def test_checkpoints(jp_fetch, contents):
    path = 'foo/a.ipynb'
    resp = await jp_fetch(
        'api', 'contents', path,
        method='GET'
    )
    model = json.loads(resp.body.decode())
    r = await jp_fetch(
        'api', 'contents', path, 'checkpoints',
        method='POST',
        allow_nonstandard_methods=True
    )
    assert r.code == 201
    cp1 = json.loads(r.body.decode())
    assert set(cp1) == {'id', 'last_modified'}
    assert r.headers['Location'].split('/')[-1] == cp1['id']

    # Modify it.
    nbcontent = model['content']
    nb = from_dict(nbcontent)
    hcell = new_markdown_cell('Created by test')
    nb.cells.append(hcell)

    # Save it.
    nbmodel = {'content': nb, 'type': 'notebook'}
    await jp_fetch(
        'api', 'contents', path,
        method='PUT',
        body=json.dumps(nbmodel)
    )

    # List checkpoints
    r = await jp_fetch(
        'api', 'contents', path, 'checkpoints',
        method='GET'
    )
    cps = json.loads(r.body.decode())
    assert cps == [cp1]

    r = await jp_fetch(
        'api', 'contents', path,
        method='GET'
    )
    nbcontent = json.loads(r.body.decode())['content']
    nb = from_dict(nbcontent)
    assert nb.cells[0].source == 'Created by test'

    # Restore Checkpoint cp1
    r = await jp_fetch(
        'api', 'contents', path, 'checkpoints', cp1['id'],
        method='POST',
        allow_nonstandard_methods=True
    )
    assert r.code == 204

    r = await jp_fetch(
        'api', 'contents', path,
        method='GET'
    )
    nbcontent = json.loads(r.body.decode())['content']
    nb = from_dict(nbcontent)
    assert nb.cells == []

    # Delete cp1
    r = await jp_fetch(
        'api', 'contents', path, 'checkpoints', cp1['id'],
        method='DELETE'
    )
    assert r.code == 204

    r = await jp_fetch(
        'api', 'contents', path, 'checkpoints',
        method='GET'
    )
    cps = json.loads(r.body.decode())
    assert cps == []


async def test_file_checkpoints(jp_fetch, contents):
    path = 'foo/a.txt'
    resp = await jp_fetch(
        'api', 'contents', path,
        method='GET'
    )
    orig_content = json.loads(resp.body.decode())['content']
    r = await jp_fetch(
        'api', 'contents', path, 'checkpoints',
        method='POST',
        allow_nonstandard_methods=True
    )
    assert r.code == 201
    cp1 = json.loads(r.body.decode())
    assert set(cp1) == {'id', 'last_modified'}
    assert r.headers['Location'].split('/')[-1] == cp1['id']

    # Modify it.
    new_content = orig_content + '\nsecond line'
    model = {
        'content': new_content,
        'type': 'file',
        'format': 'text',
    }

    # Save it.
    await jp_fetch(
        'api', 'contents', path,
        method='PUT',
        body=json.dumps(model)
    )

    # List checkpoints
    r = await jp_fetch(
        'api', 'contents', path, 'checkpoints',
        method='GET'
    )
    cps = json.loads(r.body.decode())
    assert cps == [cp1]

    r = await jp_fetch(
        'api', 'contents', path,
        method='GET'
    )
    content = json.loads(r.body.decode())['content']
    assert content == new_content

    # Restore Checkpoint cp1
    r = await jp_fetch(
        'api', 'contents', path, 'checkpoints', cp1['id'],
        method='POST',
        allow_nonstandard_methods=True
    )
    assert r.code == 204

    r = await jp_fetch(
        'api', 'contents', path,
        method='GET'
    )
    restored_content = json.loads(r.body.decode())['content']
    assert restored_content == orig_content

    # Delete cp1
    r = await jp_fetch(
        'api', 'contents', path, 'checkpoints', cp1['id'],
        method='DELETE'
    )
    assert r.code == 204

    r = await jp_fetch(
        'api', 'contents', path, 'checkpoints',
        method='GET'
    )
    cps = json.loads(r.body.decode())
    assert cps == []



import pytest
import tornado

from jupyter_server.services.contents.largefilemanager import AsyncLargeFileManager, LargeFileManager
from jupyter_server.utils import ensure_async
from ...utils import expected_http_error


@pytest.fixture(params=[LargeFileManager, AsyncLargeFileManager])
def jp_large_contents_manager(request, tmp_path):
    """Returns a LargeFileManager instance."""
    file_manager = request.param
    return file_manager(root_dir=str(tmp_path))


async def test_save(jp_large_contents_manager):
    cm = jp_large_contents_manager
    model = await ensure_async(cm.new_untitled(type='notebook'))
    name = model['name']
    path = model['path']

    # Get the model with 'content'
    full_model = await ensure_async(cm.get(path))
    # Save the notebook
    model = await ensure_async(cm.save(full_model, path))
    assert isinstance(model, dict)
    assert 'name' in model
    assert 'path' in model
    assert model['name'] == name
    assert model['path'] == path


@pytest.mark.parametrize(
    'model,err_message',
    [
        (
            {'name': 'test', 'path': 'test', 'chunk': 1},
            'HTTP 400: Bad Request (No file type provided)'
        ),
        (
            {'name': 'test', 'path': 'test', 'chunk': 1, 'type': 'notebook'},
            'HTTP 400: Bad Request (File type "notebook" is not supported for large file transfer)'
        ),
        (
            {'name': 'test', 'path': 'test', 'chunk': 1, 'type': 'file'},
            'HTTP 400: Bad Request (No file content provided)',
        ),
        (
            {'name': 'test', 'path': 'test', 'chunk': 2, 'type': 'file',
                'content': u'test', 'format': 'json'},
            "HTTP 400: Bad Request (Must specify format of file contents as 'text' or 'base64')"
        )
    ]
)
async def test_bad_save(jp_large_contents_manager, model, err_message):
    with pytest.raises(tornado.web.HTTPError) as e:
        await ensure_async(jp_large_contents_manager.save(model, model['path']))
    assert expected_http_error(e, 400, expected_message=err_message)


async def test_saving_different_chunks(jp_large_contents_manager):
    cm = jp_large_contents_manager
    model = {'name': 'test', 'path': 'test', 'type': 'file',
                'content': u'test==', 'format': 'text'}
    name = model['name']
    path = model['path']
    await ensure_async(cm.save(model, path))

    for chunk in (1, 2, -1):
        for fm in ('text', 'base64'):
            full_model = await ensure_async(cm.get(path))
            full_model['chunk'] = chunk
            full_model['format'] = fm
            model_res = await ensure_async(cm.save(full_model, path))
            assert isinstance(model_res, dict)
            assert 'name' in model_res
            assert 'path' in model_res
            assert 'chunk' not in model_res
            assert model_res['name'] == name
            assert model_res['path'] == path


async def test_save_in_subdirectory(jp_large_contents_manager, tmp_path):
    cm = jp_large_contents_manager
    sub_dir = tmp_path / 'foo'
    sub_dir.mkdir()
    model = await ensure_async(cm.new_untitled(path='/foo/', type='notebook'))
    path = model['path']
    model = await ensure_async(cm.get(path))

    # Change the name in the model for rename
    model = await ensure_async(cm.save(model, path))
    assert isinstance(model, dict)
    assert 'name' in model
    assert 'path' in model
    assert model['name'] == 'Untitled.ipynb'
    assert model['path'] == 'foo/Untitled.ipynb'